#!/usr/bin/env python

import re
import os
import sys
import stat
import errno
import subprocess
import charmhelpers.core.host
import charmhelpers.core.hookenv
import charmhelpers.fetch
from jinja2 import Environment, FileSystemLoader


required_pkgs = ['juju-core', 'juju-local', 'btrfs-tools']
hooks = charmhelpers.core.hookenv.Hooks()
config = {}


def get_juju_home(username):
    return os.path.expanduser('~%s/.juju' % username)


def get_sudoers_dir():
    return '/etc/sudoers.d'

def get_template_dir():
    d = os.path.join(charmhelpers.core.hookenv.charm_dir(), 'templates')
    return d


def juju_running():
    '''Return True if there is a bootstrapped environment.'''
    try:
        with open('/dev/null', 'w') as devnull:
            kw = {'stdout': devnull, 'stderr': devnull}
            subprocess.check_call(['juju', 'status'], **kw)
            return True
    except subprocess.CalledProcessError:
        return False


def update_sudoers_template(user):
    path = os.path.join(get_sudoers_dir(), 'juju-local')
    template = get_config_template('sudoers.tmpl')
    content = template.render({'user': user})
    with open(path, 'w') as fp:
        fp.write(content)


def update_juju_home_template(params, user, group):
    # Validate.
    assert params.get('admin_secret'), 'No secret set.'
    assert params.get('default_series'), 'No series set.'
    assert not juju_running(), 'A juju environment is bootstrapped.'

    # Populate.
    template = get_config_template('environments.yaml.tmpl')
    juju_home = get_juju_home(user)
    try:
        kw = {'owner': user, 'group': group, 'perms': 0755}
        charmhelpers.core.host.mkdir(juju_home, **kw)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    path = os.path.join(juju_home, 'environments.yaml')
    content = template.render(**params)
    charmhelpers.core.host.write_file(path, content, owner=user, group=group)


def get_config_template(name):
    template_env = Environment(loader=FileSystemLoader(get_template_dir()))
    return template_env.get_template(name)


def set_secret(config):
    if not config['admin_secret']:
        config['admin_secret'] = charmhelpers.core.host.pwgen()
        config.save()


def find_device(regex, dev='/dev'):
    '''Find the the most recent unused block device that matches the regex.'''
    pattern = re.compile(regex)
    mount = subprocess.check_output(['mount'])

    for blkdev in reversed(os.listdir(dev)):
        blkdev = os.path.join(dev, blkdev)
        if re.match(pattern, blkdev):
            # Matches the expected block device regex.

            if '%s on ' % blkdev in mount:
                # Mounted.
                continue
            if not stat.S_ISBLK(os.stat(blkdev).st_mode):
                # Not a block device.
                continue
            try:
                subprocess.check_call(['blockdev', '--rereadpt', blkdev])
            except subprocess.CalledProcessError:
                # In use.
                continue

            partition = '%s1' % blkdev
            if os.path.exists(partition):
                # Has partitions. Use the first one.
                return partition

            return blkdev


def needs_format(device, label):
    '''If the device isn't btrfs or doesn't have the same label, it needs to be
    reformatted.'''

    file_type = subprocess.check_output(['file', '-s', device])
    correct_filesystem = ' BTRFS ' in file_type
    correct_label = 'Filesystem label "%s"' % label in file_type
    return not (correct_filesystem and correct_label)


def format_device(device, label):
    subprocess.check_call(['mkfs.btrfs', '-L', label, device])


def mount_device(label, target='/var/lib/lxc'):
    # Disable device barriers and the fsync tree logging. We don't care about
    # this data on power failure, so go faster.
    entry = 'LABEL=%s\t%s\tbtrfs\tnobarrier,notreelog\t0\t0\n' % (label, target)
    with open('/etc/fstab') as fp:
        fstab = fp.read()
    if not re.match(r'\bLABEL=%s\b' % label, fstab):
        assert not re.match(r'\b%s\b' % target, fstab)
        with open('/etc/fstab', 'a') as fp:
            fp.write(entry)
    subprocess.check_call(['mount', 'LABEL=%s' % label])


def set_lxc_mount_cinder(config):
    if config['volume-ephemeral-storage']:
        return False
    if not config['volume-label']:
        return False
    if not config['volume-dev-regexp']:
        return False

    device = find_device(config['volume-dev-regexp'])
    if device:
        if needs_format(device, config['volume-label']):
            format_device(device, config['volume-label'])
        mount_device(config['volume-label'])
    return True


def set_lxc_mount_ram(config):
    size = config['lxc_in_memory']
    if size <= 0:
        return False

    # Convert from GB to bytes
    size = size * 1024 * 1024 * 1024
    uname = subprocess.check_output(['uname', '-r']).strip()
    linux_extra = 'linux-image-extra-%s' % uname
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install([linux_extra], fatal=True)
    # zram will trigger EBUSY in /sys/.../disksize if a mount happened while it
    # was loaded.
    subprocess.call(['rmmod', 'zram'])
    subprocess.check_output(['modprobe', 'zram'])
    with open('/sys/block/zram0/disksize', 'w') as fp:
        fp.write(str(size))

    device = '/dev/zram0'
    label = 'var-lib-lxc'
    if needs_format(device, label):
        format_device(device, label)
    # FIXME do not set up a permanent mountpoint. zram devices unsurprisingly
    # do not survive an rmmod.
    mount_device(label)
    return True


def setup_caching_proxy(config):
    '''Setup squid-deb-proxy for caching packages.'''
    if not config['caching_proxy']:
        return
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(['squid-deb-proxy'], fatal=True)
    path = '/etc/squid-deb-proxy/mirror-dstdomain.acl.d/20-juju-local'
    content = 'ppa.launchpad.net\nprivate-ppa.launchpad.net\n'
    charmhelpers.core.host.write_file(path, content)
    svc = 'squid-deb-proxy'
    if not charmhelpers.core.host.service_restart(svc):
        charmhelpers.core.host.service_start(svc)


@hooks.hook('user-relation-changed')
@hooks.hook('config-changed')
def config_changed():
    # Populate.
    set_secret(config)
    user = charmhelpers.core.hookenv.relation_get('username')
    group = charmhelpers.core.hookenv.relation_get('group')
    if user and group:
        update_juju_home_template(config, user, group)
        update_sudoers_template(user)
    if not set_lxc_mount_ram(config):
        set_lxc_mount_cinder(config)
    setup_caching_proxy(config)

@hooks.hook('upgrade-charm')
@hooks.hook()
def install():
    # Install packages.
    charmhelpers.core.hookenv.log('Installing packages.')
    charmhelpers.fetch.configure_sources(update=True)
    charmhelpers.fetch.apt_install(required_pkgs, fatal=True)
    config_changed()


if __name__ == '__main__':
    config = charmhelpers.core.hookenv.config()
    hooks.execute(sys.argv)
