import os
import unittest
import hooks
import mock
import shutil
import tempfile
import subprocess
import grp


PATCHES = [
    'charmhelpers.core.hookenv.log',
    'charmhelpers.core.host.log',
]

class HookTestCase(unittest.TestCase):
    def setUp(self):
        super(HookTestCase, self).setUp()
        for spec in PATCHES:
            attr_name = spec.split('.')[-1]
            _m = mock.patch(spec)
            setattr(self, attr_name, _m.start())
            self.addCleanup(_m.stop)        

class TestUtilityFunctions(HookTestCase):
    def test_juju_running(self):
        kw = {'returncode': 1, 'cmd': 'juju status', 'output': 'Failed.'}
        e = subprocess.CalledProcessError(**kw)
        with mock.patch('subprocess.check_call', side_effect=e):
            self.assertFalse(hooks.juju_running())
        with mock.patch('subprocess.check_call', return_value=None):
            self.assertTrue(hooks.juju_running())


class TestTemplatePopulation(HookTestCase):
    def setUp(self):
        super(TestTemplatePopulation, self).setUp()
        env = {'CHARM_DIR': '.', 'JUJU_UNIT_NAME': 'juju-local/0'}
        patch_env = mock.patch.dict('os.environ', env)
        patch_env.start()
        self.addCleanup(patch_env.stop)

        _p = mock.patch('hooks.juju_running', return_value=False)
        _p.start()
        self.addCleanup(_p.stop)

        self.juju_home = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.juju_home)
        _p = mock.patch('hooks.get_juju_home', return_value=self.juju_home)
        _p.start()
        self.addCleanup(_p.stop)

    def test_juju_population(self):
        secret = 'A'
        series = 'precise'
        params = {'admin_secret': secret, 'default_series': series}
        user = os.environ['USER']
        group = grp.getgrgid(os.getgroups()[0]).gr_name
        
        hooks.update_juju_home_template(params, user, group)
        with open(os.path.join(self.juju_home, 'environments.yaml')) as fp:
            tmpl_data = fp.read()
        self.assertIn('admin-secret: %s' % secret, tmpl_data)
        self.assertIn('series: %s' % series, tmpl_data)

    def test_sudoers_population(self):
        user = os.environ['USER']
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, temp_dir)
        with mock.patch('hooks.get_sudoers_dir', return_value=temp_dir):
            hooks.update_sudoers_template(user)
        with open(os.path.join(temp_dir, 'juju-local')) as fp:
            tmp_data = fp.read()
        self.assertIn('%s ALL=(root) NOPASSWD: ALL' % user, tmp_data)


class TestVolumeSupport(HookTestCase):
    def test_find_device(self):
        temp_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, temp_dir)
        for blkdev in ('vda', 'vdb', 'vdc', 'zero'):
            with open(os.path.join(temp_dir, blkdev), 'w'):
                pass

        _p = mock.patch('stat.S_ISBLK', return_value=True)
        _p.start()
        self.addCleanup(_p.stop)

        _p = mock.patch('subprocess.check_call')
        _p.start()
        self.addCleanup(_p.stop)

        pattern = os.path.join(temp_dir, 'vd[b-z]')
        result = hooks.find_device(pattern, dev=temp_dir)
        self.assertEqual(result, os.path.join(temp_dir, 'vdc'))


if __name__ == '__main__':
    unittest.main()
